@@ -6,12 +6,17 @@ Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
 
 from __future__ import division
 
+import abc
 import collections
+import functools
+import hashlib
+import io
 import logging
 import math
 import os.path
+import sys
 
-logger = logging.getLogger('__name__')
+logger = logging.getLogger(__name__)
 # logger.setLevel('DEBUG')
 
 
@@ -24,6 +29,8 @@ class Store(object):
 
     """
 
+    __metaclass__ = abc.ABCMeta
+
     ignoreExtraVolumes = False
 
     def __init__(self, userPath, isDest, dryrun):
@@ -54,6 +61,12 @@ class Store(object):
 
         self.dryrun = dryrun
 
+    def listContents(self):
+        """ Return list of volumes or diffs in this Store's selected directory. """
+        vols = list(self.listVolumes())
+        vols.sort(key=lambda v: self.getSendPath(v))
+        return [vol.display(self, detail="line") for vol in vols]
+
     def listVolumes(self):
         """ Return list of all volumes in this Store's selected directory. """
         for (vol, paths) in self.paths.items():
@@ -88,10 +101,14 @@ class Store(object):
 
         """
         logger.debug("%s", paths)
+        if not paths:
+            path = os.path.basename(self.userPath) + '/Anon'
         try:
-            return self._fullPath([p for p in paths if not p.startswith("/")][0])
+            path = [p for p in paths if not p.startswith("/")][0]
         except IndexError:
-            return self._fullPath(os.path.basename(list(paths)[0]))
+            path = os.path.basename(list(paths)[0])
+
+        return self._fullPath(path)
 
     def _fullPath(self, path):
         if path.startswith("/"):
@@ -115,35 +132,56 @@ class Store(object):
         else:
             return fullPath
 
-    def _skipDryRun(self, logger):
-        return skipDryRun(logger, self.dryrun)
+    def _skipDryRun(self, logger, level='DEBUG', dryrun=None):
+        return skipDryRun(logger, dryrun or self.dryrun, level)
 
     # Abstract methods
 
+    @abc.abstractmethod
     def _fillVolumesAndPaths(self):
         """ Fill in self.paths. """
         raise NotImplementedError
 
+    @abc.abstractmethod
     def getEdges(self, fromVol):
         """ Return the edges available from fromVol. """
         raise NotImplementedError
 
+    @abc.abstractmethod
     def hasEdge(self, diff):
         """ True if Store already contains this edge. """
         raise NotImplementedError
 
+    @abc.abstractmethod
     def receive(self, diff, paths):
         """ Return Context Manager for a file-like (stream) object to store a diff. """
         raise NotImplementedError
 
-    def send(self, diff, streamContext, progress=True):
-        """ Write the diff (toVol from fromVol) to the stream context manager. """
+    @abc.abstractmethod
+    def send(self, diff, progress=True):
+        """ Return Context Manager for a file-like (stream) object to send a diff. """
         raise NotImplementedError
 
+    @abc.abstractmethod
     def receiveVolumeInfo(self, paths):
         """ Return Context Manager for a file-like (stream) object to store volume info. """
         raise NotImplementedError
 
+    @abc.abstractmethod
+    def keep(self, diff):
+        """ Mark this diff (or volume) to be kept in path. """
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def deleteUnused(self):
+        """ Delete any old snapshots in path, if not kept. """
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def deletePartials(self):
+        """ Delete any old partial uploads/downloads in path. """
+        raise NotImplementedError
+
 
 class Diff:
 
@@ -185,6 +223,74 @@ class Diff:
         self._updateSize()
         return self._sizeIsEstimated
 
+    def sendTo(self, dest, chunkSize, progress=True):
+        """ Send this difference to the dest Store. """
+        logger.info("%s: %s", "Keep" if self.sink == dest else "Xfer", self)
+
+        vol = self.toVol
+        paths = self.sink.getPaths(vol)
+
+        if self.sink == dest:
+            self.sink.keep(self)
+            return
+
+        streamContext = dest.receive(self, paths)
+
+        sendContext = self.sink.send(self, progress)
+
+        # try:
+        #     streamContext.metadata['btrfsVersion'] = self.btrfsVersion
+        # except AttributeError:
+        #     pass
+
+        try:
+            streamContext.progress = progress
+        except AttributeError:
+            pass
+
+        try:
+            chunkSize = streamContext.chunkSize
+        except AttributeError:
+            pass
+
+        if sendContext is not None and streamContext is not None:
+            with streamContext as writer:
+                # Open reader after writer,
+                # so any raised errors will abort write before writer closes.
+                with sendContext as reader:
+                    checkBefore = None
+                    if hasattr(writer, 'skipChunk'):
+                        checkBefore = hasattr(reader, 'checkSum')
+
+                    while True:
+                        if checkBefore is True:
+                            (size, checkSum) = reader.checkSum(chunkSize)
+
+                            if writer.skipChunk(size, checkSum):
+                                reader.seek(size, io.SEEK_CUR)
+                                continue
+
+                        data = reader.read(chunkSize)
+                        if len(data) == 0:
+                            break
+
+                        if checkBefore is False:
+                            checkSum = hashlib.md5(data).hexdigest()
+
+                            if writer.skipChunk(len(data), checkSum, data):
+                                continue
+
+                        writer.write(data)
+
+        if vol.hasInfo():
+            infoContext = dest.receiveVolumeInfo(paths)
+
+            if infoContext is None:
+                vol.writeInfo(sys.stdout)
+            else:
+                with infoContext as stream:
+                    vol.writeInfo(stream)
+
     def _updateSize(self):
         if self._size and not self._sizeIsEstimated:
             return
@@ -199,12 +305,12 @@ class Diff:
 
     def __str__(self):
         """ human-readable string. """
-        return u"%s from %s (%s%s) in %s" % (
+        return u"%s from %s (%s%s)" % (
             self.toVol.display(self.sink),
-            self.fromVol.display(self.sink) if self.fromVol else "<None>",
+            self.fromVol.display(self.sink) if self.fromVol else "None",
             "~" if self.sizeIsEstimated else "",
             humanize(self.size),
-            self.sink,
+            # self.sink,
         )
 
 
@@ -236,12 +342,16 @@ class Volume:
     def writeInfo(self, stream):
         """ Write information about diffs into a file stream for use later. """
         for (fromUUID, size) in Diff.theKnownSizes[self.uuid].iteritems():
-            if size is not None and fromUUID is not None:
-                stream.write(str("%s\t%s\t%d\n" % (
-                    self.uuid,
-                    fromUUID,
-                    size,
-                )))
+            if size is None or fromUUID is None:
+                continue
+            if not isinstance(size, int):
+                logger.warning("Bad size: %s", size)
+                continue
+            stream.write(str("%s\t%s\t%d\n" % (
+                self.uuid,
+                fromUUID,
+                size,
+            )))
 
     def hasInfo(self):
         """ Will have information to write. """
@@ -257,6 +367,12 @@ class Volume:
         """ Read previously-written information about diffs. """
         for line in stream:
             (toUUID, fromUUID, size) = line.split()
+            try:
+                size = int(size)
+            except:
+                logger.warning("Bad size: %s", size)
+                continue
+            logger.debug("diff info: %s %s %d", toUUID, fromUUID, size)
             Diff.theKnownSizes[toUUID][fromUUID] = size
 
     def __unicode__(self):
@@ -276,7 +392,10 @@ class Volume:
 
     def display(self, sink=None, detail='phrase'):
         """ Friendly string for volume, using sink paths. """
-        if detail in ('line', 'paragraph') and self.size is not None:
+        if not isinstance(detail, int):
+            detail = detailNum[detail]
+
+        if detail >= detailNum['line'] and self.size is not None:
             size = " (%s%s)" % (
                 printBytes(self.size),
                 "" if self.exclusiveSize is None else (
@@ -287,7 +406,7 @@ class Volume:
             size = ""
 
         vol = "%s %s" % (
-            printUUID(self._uuid),
+            printUUID(self._uuid, detail - 1),
             sink.getSendPath(self) if sink else "",
         )
 
@@ -303,6 +422,17 @@ class Volume:
         else:
             return cls(vol)
 
+detailTypes = ('word', 'phrase', 'line', 'paragraph')
+detailNum = {t: n for (n, t) in zip(xrange(len(detailTypes)), detailTypes)}
+
+
+def display(obj, detail='phrase'):
+    """ Friendly string for volume, using sink paths. """
+    try:
+        return obj.display(detail=detail)
+    except AttributeError:
+        return str(obj)
+
 
 def printBytes(number):
     """ Return a human-readable string for number. """
@@ -323,16 +453,30 @@ def humanize(number):
     return "%.4g %s" % (mantissa, units[pow])
 
 
-def printUUID(uuid):
+def printUUID(uuid, detail='word'):
     """ Return friendly abbreviated string for uuid. """
+    if not isinstance(detail, int):
+        detail = detailNum[detail]
+
+    if detail > detailNum['word']:
+        return uuid
+
     if uuid is None:
         return None
+
     return "%s...%s" % (uuid[:4], uuid[-4:])
 
 
-def skipDryRun(logger, dryRun):
-    """ Print or log command about to be run, and return True if should be skipped. """
-    return _skipRun if dryRun else logger.debug
+def skipDryRun(logger, dryRun, level=logging.DEBUG):
+    """ Print or log command about to be run.
+
+    Return True if should be skipped. """
+    # This is an undocumented "feature"
+    # logging.log() require a numeric level
+    # logging.getLevelName() also maps names to numbers
+    if not isinstance(level, int):
+        level = logging.getLevelName(level)
+    return _skipRun if dryRun else functools.partial(logger.log, level)
 
 
 def _skipRun(format, *args):
